#!/usr/bin/perl -w
#
#-----------------------------------------------------------------------
# s25rtarpitgreylist.pl   ver 1.3.1  (2008/12/2)
# 
#   This program carries out this program right before it is carried out 
#   smtp program (qmail-smtpd) of qmail and refuses the connection of 
#   the host like the origin of spam transmission of a message.
#
#   This program uses algorithm to demand an email retransmission called 
#   greylisting from the judgment of a host refusing it. 
#   ( The spammer wants to transmit an email in large quantities early. 
#     Therefore, the spammer hates retransmission. As a result, 
#     the spammer does not transmit a message again.                  )
#
#   However, this program uses algorithm called S25R (selective SMTP refusal) 
#   because a side effect is too big only in algorithm called this greylist.
#   And only a host like the dynamic IP address that spammers use 
#   well of greylisting intend for it. 
#
#   In addition, this program refuses the connection of the origin of 
#   spam transmission of a message host by carrying out tarpit ( it delay 
#   a reply at the time of SMTP connection intentionally).
#   ( The spammer wants to transmit an email in large quantities early. 
#     As a result, spammer hate a reply delay and the spammer cuts off
#     connection by oneself                                           )
#
#----------------------------------------------------------------------
# Copyright (C) 2008  chicchi    http://chiji.atnifty.com/  
#                                mailto:webmaster@chiji.atnifty.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
#-----------------------------------------------------------------------
# Because the author does not assume English a native language, 
# with "Yahoo! Translation"(http://honyaku.yahoo.co.jp/transtext)  of Yahoo! japan and
# "excite translation"(http://www.excite.co.jp/world/) of Excite Japan, 
# I translate Japanese into English automatically.
# Therefore there is the point where a sentence is unnatural. Please understand it.
#
#
#---------------------------------------#
# use    (do not edit)                  #
#---------------------------------------#
use strict;
use Sys::Syslog qw(:DEFAULT setlogsock);
use Time::Local;


#---------------------------------------#
# const  (do not edit)                  #
#---------------------------------------#

# log level
use constant LOG_LV_DEBUG => 0;
use constant LOG_LV_INFO => 1;
use constant LOG_LV_FORCE => 2;

# log type
use constant LOG_INFO => 'INFO';
use constant LOG_WARN => 'WARN';
use constant LOG_ERR => 'ERR';
use constant LOG_NG => 'NG';
use constant LOG_OK => 'OK';

#---------------------------------------#
# setting                               #
#---------------------------------------#

# ---- general ----
# path of data directory
my ${Dir_Data}="/var/qmail/s25rtarpitgreylist";

# qmail control directory
my ${Dir_Qmailcontrol}="/var/qmail/control";

# path of temporary directory
#     This directory needs authority to be able to write in qmaild at.
my ${Dir_Tmp}="${Dir_Data}/tmp";

# add ver1.2.3
# mode use flock
#  0 - disable flock. use lockfile. Lower than perl version 5.004 are this one.
#  1 - enable flock. it does not use lockfile.(defaut)
my ${Enable_Flock}=1;

# add ver1.3
# This is an option displaying a more accurate reason.
# 0 - disable (default)
# 1 - enable
my ${Mode_PrintMsg_AccurateReason}=0;

# add ver1.3.1
# mode use eval
#  0 - disable eval.
#  1 - enable eval.(defaut)
my ${Enable_Eval}=1;

# ---- Setting of the log output ----
# set a level of the log output
#  LOG_LV_DEBUG  - output debug and infomation log
#  LOG_LV_INFO   - output infomation log only
my ${Mode_Log_Output_Level} = LOG_LV_DEBUG;

# ---- whitelist ----
# path of whitelist file of ip addr
my ${Path_Whitelist} = "${Dir_Data}/whitelist"; 

# path of whitelist file 
my ${Path_Whitelist_Hostname}="${Dir_Data}/whitelist_hostname";


# ---- blacklist -----
# Mode_Blacklist
#   0 - disable blacklist
#   1 - s25r before blacklist (It gives priority to s25r) (default)
#   2 - blacklist before s25r (It gives priority to blacklist)
my ${Mode_Blacklist} = 1;

# path of blacklist file of ip addr
my ${Path_Blacklist} = "${Dir_Data}/blacklist"; 

# path of blacklist file 
my ${Path_Blacklist_Hostname}="${Dir_Data}/blacklist_hostname";


# ---- greylisting -----
# Mode_Greylist
#   0 - disable Greylist
#   1 - enable Greylist (default)
my ${Mode_Greylist} = 1;

# path of greylist file
my ${Path_Greylist} = "${Dir_Data}/greylist"; 

# path of greylist lock file
#     The directory arranging this file needs authority to be able to write in qmaild at.
my ${Path_Lock_Greylist} = "${Dir_Tmp}/greylist.lock"; 

# path of greylist tmp file
#     The directory arranging this file needs authority to be able to write in qmaild at.
my ${Path_Tmp_Greylist} = "${Dir_Tmp}/greylist.tmp"; 


# path of file which manages last deletion time
my ${Path_Lastdeletetime_Greylist} = "${Dir_Data}/lastdeletetime";

# interval to check greylist（second） (default:30*60)
my ${Sec_Checkinterval_Greylist}=30*60;

# Number of seconds suppression (default:2*60)
my ${Sec_Suppression_Connect} = 2 * 60;

# Number of seconds validity of a host admitting access  (default:32 * 24 * 60 * 60)
my ${Sec_Validity_Goodhost}= 32 * 24 * 60 * 60;

# Number of seconds availability of the host who makes access suppression in greylist (default:24 * 60 * 60)
my ${Sec_Validity_Nogoodhost}=24 * 60 * 60;

# message to display at the time of connection completion 
#       ( if no set,smtpgreeting or me in /var/qmail/control )
# Ex:my ${Msg_Connection}="hogehoge.com";
my ${Msg_Connection}="";

#  Number of seconds SMTP timeout (default:5 * 60)
my ${Sec_Smtptimeout}= 5 * 60 ;

# Number of seconds reply delay between commands (default:1)
my ${Sec_Smtpcmdwait}=1;

# Threshold Count of Come soon
# 0 -- disable
# 1< -- When the number of times more than this set value came, I refuse it
my ${Threshold_Count_Comesoon}=1;

# The setting that it investigates IP file to the third octet (default:0)
# 0 - It is an investigation object in all parts of the IP address
# 1 - It does the investigation object of the IP address by the third octet
my ${Mode_Ipfile_Thirdoctet}=0;

# add ver1.2
# The number of the SMTP command practice to refuse registration to greylist. 
# When a value to set is equal to or less than this price, 
# it is refused registration to greylist.
#   QUIT only or HELO -> QUIT only
# 0 -- disable
# 1< -- enable   (2 -- default) 
my ${Threshold_Count_SmtpCommand}=2;

# add ver1.2
# Please set the number of the unexpected SMTP commands to refuse connection.
# 0 -- disable (default)
# 1< -- enable 
my ${Threshold_Count_UnknownSmtpCmd}=0;

# add ver1.2
# When the connection from an origin of spam transmission of 
# a message host did timeout, this program cancels registration to
# greylist of the host.
# 0 -- disable (default)
# 1 -- enable
my ${Mode_SmtpTimeout_Deny}=0;

# add ver1.2.2
#
my ${Count_Retry_GreylistLock}=2;


# ---- S25R ----
# Mode_S25rlist
#   0 - disable S25rlist
#   1 - enable S25rlist  (default)
my ${Mode_S25rlist}= 1;
# location for s25r file of hostname
my ${Path_S25rlist_Hostname}="${Dir_Data}/s25rlist_hostname";

# ---- tarpit ----
# I perform tarpit before demanding retransmission.
# I accept the host if connected even if I do tarpit.
# This is one of the salvation to reduce a work load to add the SMTP
# server that is the justice that cannot transmit a message again to
# whitelist.
# (this uses http://honyaku.yahoo.co.jp/transtext, and it is the sentence
# that it converted into English from Japanese)
#  Do you perform tarpit?
#    0 -- No   (I am the same as movement of normal qgreylist)
#    1 -- yes  (I carry out tarpit before demanding retransmission from the host whom I do not know ) (default)
#    2 -- yes  (I carry out tarpit for all hosts before checking greylist. )
my ${Mode_Tarpit} = 1;
#   tarpit second (default:65   If you want to make it severer, 125 is recommended  )
my ${Sec_Tarpit} = 65;
#   If I do tarpit and do not disconnect, I receive an email
#    0 -- after 450  (default)
#    1 -- accept  
my ${Mode_Accept_Tarpitwaithost} = 0;

# ---- bad helo host ----
# processing method when I gave badhelohost by an HELO command
#   0 -- greylisting(after 450)
#   1 -- deny(550) (default)
my ${Mode_Badhelohost}=1;
# location for blacklist file of helo_host
my ${Path_Blacklist_Helohost}="${Dir_Data}/blacklist_helohost";

# ---- compare helohost and connect host ----
# This function compares HELO host name with connection host name and 
# cancels greylist registration when both are different.
# This function is experimental. You had better invalidate it.
# 0 -- disable        (default)
# 1 -- TLD only
# 2 -- TLD and SLD        
# 3 -- TLD and SLD and ???
my ${Mode_Compare_Hostname_Octetlevel} = 0;

# ver1.1 add
# ---- compare check remote ip if helo is ip address ----
# When a connection host gave an IP address in HELO, this program checks 
# the IP address of the connection host, and this program cuts off
# connection if different.
#   0 - disable 
#   1 - enable  (default)
my ${Mode_Check_RemotoIp_HeloIp} = 1;

# ---- smtp auth escape----
# Do you register a SMTP-AUTH host with greylist?
#    0 -- no    (default)
#    1 -- yes
#   It is very high-risk to validate this setting. 
#   Because the reason is because the connection from 
#   the host is admitted unconditionally when carried out an AUTH command. 
#    （A spammer may carry out the AUTH command of the dummy）
my ${Mode_Escape_SmtpAuth}=0;

# ver1.3.0 add
# This is setting to escape SMTP-AUTH connection. （only qmail-vida)
#    0 -- disable. no escape. (defalut)
#    1 -- enable. escape SMTP-AUTH connection. 
#  But it is necessary to start qmail-smtpd from s25rtarpitgreylist.pl. 
#   （ usually start qmail-smtpup from s25rtarpitgreylist.pl.）
#   It is necessary to set it ,${Threshold_Count_SmtpCommand}=0 or 1 then.
my ${Mode_Escape_SmtpAuth_ForVida} = 0;


#---------------------------------------#
# grobal varieble   (do not edit)       #
#---------------------------------------#
my ${Timesirial_Access}="";
my ${Timesirial_Create}="";
my ${Remoteip} = "-";
my ${Remotehostip} = "";
my ${Objectipaddr} = "";
my ${Count_Smtpcmd}=0;


#---------------------------------------#
# setting proc   (do not edit)          #
#---------------------------------------#
# This acquires the domain name of the server of own.
my ${Domain}="hoge.com";
if ( -r "${Dir_Qmailcontrol}/smtpgreeting" ) {
	open DOMAIN,"${Dir_Qmailcontrol}/smtpgreeting";
	${Domain}=<DOMAIN>;
	close DOMAIN;
}elsif( -r "${Dir_Qmailcontrol}/me") {
	open DOMAIN,"${Dir_Qmailcontrol}/me";
	${Domain}=<DOMAIN>;
	close DOMAIN;
}
chomp ${Domain};
if ( ( ! defined ${Msg_Connection} ) or ( ${Msg_Connection} eq '' ) ) {
	${Msg_Connection}=${Domain};
} 


# syslog
setlogsock 'unix';
openlog "greylist", "pid", "mail";
#openlog "s25rtarpitgreylist", "pid", "mail";
# It does not buffer the output
$|=1;

# use flock
if ( ${Enable_Flock} == 1) {
	use Fcntl qw(:DEFAULT :flock);
}

#---------------------------------------#
# sub define    (do not edit)           #
#---------------------------------------#
sub write_log {
	my ${level_arg};
	my ${type_arg};
	my ${msg_arg};
	my ${funcname_arg};
	my ${sysloglevel};
	
	my ${remotehostip_output};
	my ${remoteip_output};
	
	( ${level_arg} , ${type_arg} , ${funcname_arg} , ${msg_arg} ) = @_;
	
	if ( ${level_arg} >= ${Mode_Log_Output_Level} ) {
		if ( ${level_arg} == LOG_LV_DEBUG ) {
			${sysloglevel}="debug";
		}elsif ( ${level_arg} == LOG_LV_INFO ) {
			${sysloglevel}="info";
		}elsif ( ${level_arg} == LOG_LV_FORCE ) {
			${sysloglevel}="err";
		}
		
		if ( ( ! defined ${Remotehostip} ) or ( ${Remotehostip} eq '' ) ) {
			${remotehostip_output} = '-';
		}else{
			${remotehostip_output} = ${Remotehostip};
		}
		
		if ( ( ! defined ${Remoteip} ) or ( ${Remoteip} eq '' ) ) {
			${remoteip_output} = '-';
		}else{
			${remoteip_output} = ${Remoteip};
		}
		# log format
		# hostname (ip_address) INFO/ERR/WARN/NG/OK : function_name : reasons etc
		syslog "${sysloglevel}","%s (%s) %s : %s : %s","${remotehostip_output}","${remoteip_output}",${type_arg},${funcname_arg},${msg_arg};
	}
}

sub sigexit_abort {
	write_log LOG_LV_FORCE,LOG_ERR,"-","ABORT";

	print "451 Temporary local problem. Please try later.\r\n";
	exit;
}
$SIG{'ABRT'} = "sigexit_abort";


sub conv_logtime{
	my ${gettime};
	my ${sec};
	my ${min};
	my ${hour};
	my ${mday};
	my ${mon};
	my ${year};

	( ${gettime} ) = @_;
	( ${sec}, ${min}, ${hour}, ${mday}, ${mon}, ${year}) = localtime ${gettime};
	return  sprintf "%04d-%02d-%02d %02d:%02d", ${year}+1900, ${mon}+1,${mday}, ${hour}, ${min}, ${sec};
}


sub create_file {
	my ${file};
	my ${nowtime};

	( ${file} ) = @_;

	if ( ! open FILE, "> ${file}" ) {
		write_log LOG_LV_FORCE,LOG_ERR,"-","failed to create file : ${file}";
		print "450 temporary problem \r\n";
		exit;
	}
	close FILE;

	${nowtime}= time;
	utime ${nowtime},${nowtime},${file};
}

# ---- greylist -----

sub remove_greylistlock {
	if ( -f ${Path_Lock_Greylist} ) {
		if ( ! unlink ${Path_Lock_Greylist} ) {
			write_log LOG_LV_FORCE,LOG_WARN,"-","failed to unlink greylist lockfile: ${Path_Lock_Greylist}";
		}
	}
}

sub sigexit_check_greylistlock_timeout {
	write_log LOG_LV_INFO,LOG_ERR,"greylisting","greylist lock timeout";

	print "451 Temporary local problem. Please try later.\r\n";
	exit;
}


sub check_greylistlock {
	my ${time_createlock}; # add ver1.2.2
	my ${timestamp_createlock}; # add ver1.2.2
	
	$SIG{'ALRM'} = "sigexit_check_greylistlock_timeout";
	
	alarm ${Sec_Smtptimeout};
	while ( -f ${Path_Lock_Greylist} ){
		# add ver1.2.2 
		${time_createlock} = (stat ${Path_Lock_Greylist})[9];
		if (  ( defined ${time_createlock} ) and ( ${time_createlock} < ( time - ( ${Sec_Smtptimeout} + 10 ) ) ) ){
			remove_greylistlock;
			${timestamp_createlock} = conv_logtime ${time_createlock};
			write_log LOG_LV_INFO,LOG_WARN,"greylisting","delete lock file  ( may be the made wreckage ${timestamp_createlock} )" ;
			last;
		}
		
		sleep 1;
	}
	alarm 0;
	$SIG{'ALRM'} = "";
}

sub create_greylistlock {
	my ${count_fail} = 0 ;	# add ver1.2.2 
	
	while ( ${count_fail} <= ${Count_Retry_GreylistLock} ) {	# add ver1.2.2 
		if ( ! open FILE,"> ${Path_Lock_Greylist}") {
			if ( ! -f ${Path_Lock_Greylist} ) {	# add ver1.2.2 
				write_log LOG_LV_FORCE,LOG_ERR,"-","failed to create greylist lock file : ${Path_Lock_Greylist}";
				# modify ver1.3.0
				#print "450 temporary problem \r\n";
				print "450 Temporary local problem. Please try later. \r\n";
				exit;
			}else{
				# add ver1.2.2 
				write_log LOG_LV_FORCE,LOG_WARN,"-","waited to create greylist lock file ( open lock ): ${Path_Lock_Greylist}";
				check_greylistlock;
				${count_fail}++;
				next;
			}
		} else {
			close FILE;
			# modify ver1.2.4
			# return ;
			return undef;
		}
	}
	
	# add ver1.2.2 
	# fail 
	write_log LOG_LV_FORCE,LOG_ERR,"-","failed to create greylist lock file ( long open lock ): ${Path_Lock_Greylist}";
	# modify ver1.3.0
	#print "450 temporary problem \r\n";
	print "450 Temporary local problem. Please try later. \r\n";
	exit;
}

# add ver1.2.3
sub remove_greylisttmp {
	if ( -f ${Path_Tmp_Greylist} ) {
		if ( ! unlink ${Path_Tmp_Greylist} ) {
			write_log LOG_LV_FORCE,LOG_WARN,"-","failed to unlink greylist tmpfile: ${Path_Tmp_Greylist}";
		}
	}
}

sub sigexit_check_greylisttmp_timeout {
	write_log LOG_LV_INFO,LOG_ERR,"greylisting","greylist tmp timeout";

	print "451 Temporary local problem. Please try later.\r\n";
	exit;
}

sub check_greylisttmp {
	my ${time_createtmp}; # add ver1.2.2
	my ${timestamp_createtmp}; # add ver1.2.2
	
	$SIG{'ALRM'} = "sigexit_check_greylisttmp_timeout";
	
	alarm ${Sec_Smtptimeout};
	while ( -f ${Path_Tmp_Greylist} ){
		${time_createtmp} = (stat ${Path_Tmp_Greylist})[9];
		if (  ( defined ${time_createtmp} ) and ( ${time_createtmp} < ( time - ( ${Sec_Smtptimeout} + 10 ) ) ) ){
			remove_greylisttmp;
			${timestamp_createtmp} = conv_logtime ${time_createtmp};
			write_log LOG_LV_INFO,LOG_WARN,"greylisting","delete tmp file  ( may be the made wreckage ${timestamp_createtmp} )" ;
			last;
		}
		
		sleep 1;
	}
	alarm 0;
	$SIG{'ALRM'} = "";
}

sub copy_file_noflock {
	my ${from}="";
	my ${to}="";
	my ${getline}="";

	( ${from},${to} )=@_;

	if ( ! -r ${from} ) {
		write_log LOG_LV_FORCE,LOG_ERR,"-","Can't read file (copy): ${from}";
		print "450 temporary problem \r\n";
		exit;	
	}

	if ( open FH_FROM,${from} ) {
		if ( open FH_TO,">${to}" ) {
			while ( ${getline} = <FH_FROM> ) {
				printf FH_TO "%s",${getline};
			}

			close FH_TO;
		}else{
			write_log LOG_LV_FORCE,LOG_ERR,"-","failed to open file write mode (copy): ${to} ";
			print "450 temporary problem \r\n";
			exit;
		}

		close FH_FROM;
	}else{
		write_log LOG_LV_FORCE,LOG_ERR,"-","failed to open file (copy): ${from} ";
		print "450 temporary problem \r\n";
		exit;	
	}
}

# add ver1.2.3
sub copy_file_flock {
	my ${from}="";
	my ${to}="";
	my ${getline}="";

	( ${from},${to} )=@_;

	if ( ! -r ${from} ) {
		write_log LOG_LV_FORCE,LOG_ERR,"-","Can't read file (copy): ${from}";
		print "450 temporary problem \r\n";
		exit;	
	}

	# set alarm
	$SIG{'ALRM'} = "sigexit_check_greylistlock_timeout";
	alarm ${Sec_Smtptimeout};
	
	if ( sysopen(FH_FROM,${from},O_RDONLY) ) {
		flock FH_FROM,LOCK_EX;
		if ( sysopen FH_TO,${to},O_RDWR|O_CREAT|O_SYNC ) {
			flock FH_TO,LOCK_EX;
			
			truncate( FH_TO,0 );
			seek(FH_TO,0,0);
			while ( ${getline} = <FH_FROM> ) {
				printf FH_TO "%s",${getline};
			}

			close FH_TO;
		}else{
			write_log LOG_LV_FORCE,LOG_ERR,"-","failed to open file write mode (copy): ${to} ";
			print "450 temporary problem \r\n";
			exit;
		}

		close FH_FROM;
	}else{
		write_log LOG_LV_FORCE,LOG_ERR,"-","failed to open file (copy): ${from} ";
		print "450 temporary problem \r\n";
		exit;	
	}
	
	# unset alarm
	alarm 0;
	$SIG{'ALRM'} = "";
}

sub copy_file {
	my ${from}="";
	my ${to}="";

	( ${from},${to} )=@_;
	
	if ( ${Enable_Flock} == 1 ) {
		copy_file_flock ${from},${to};
	}else{
		copy_file_noflock ${from},${to};
	}

}

# add ver1.2.3
sub copy_file_FH {
	my $FH_FROM;
	my $FH_TO;
	my ${getline} ="";
	
	( $FH_FROM , $FH_TO ) = @_;
	
	if ( ! truncate( $FH_TO,0 ) ) {
		# modify ver1.2.4
		# return ;
		return undef;
	}
	if ( ! seek($FH_TO,0,0) ) {
		# modify ver1.2.4
		# return ;
		return undef;
	}
	
	if ( ! seek($FH_FROM,0,0) ) {
		# modify ver1.2.4
		# return ;
		return undef;
	}
		
	
	while ( ${getline} = <$FH_FROM> ){
		if ( ! printf $FH_TO "%s",${getline} ) {
			# modify ver1.2.4
			# return ;
			return undef;
		}
	}
	return 1;
}

sub get_create_access_time_for_data {
	my ${inputdata}="";

	my ${ipaddr_line}="";
	my ${year_create_line}="";
	my ${month_create_line}="";
	my ${date_create_line}="";
	my ${hour_create_line}="";
	my ${minute_create_line}="";
	my ${second_create_line}="";
	my ${year_access_line}="";
	my ${month_access_line}="";
	my ${date_access_line}="";
	my ${hour_access_line}="";
	my ${minute_access_line}="";
	my ${second_access_line}="";
	my ${count_comesoon}="";
	my ${timesirial_create};
	my ${timesirial_access};
	my ${flag_invalidformat}=0; # add ver1.2.1
	
	# add ver1.2.1
	my ${year_create_tmp}="";
	my ${month_create_tmp}="";
	my ${date_create_tmp}="";
	my ${hour_create_tmp}="";
	my ${minute_create_tmp}="";
	my ${second_create_tmp}="";
	my ${year_access_tmp}="";
	my ${month_access_tmp}="";
	my ${date_access_tmp}="";
	my ${hour_access_tmp}="";
	my ${minute_access_tmp}="";
	my ${second_access_tmp}="";
	my ${count_comesoon_tmp}="";


	( ${inputdata} ) = @_;

	if ( ( defined ${inputdata} ) and (  ${inputdata} ne '' ) ) {
		# When it was able to acquire data, it resolves data and analyzes it.
		${inputdata} =~ m/(.+)\s+([0-9]+)\-([0-9]+)\-([0-9]+)\s([0-9]+)\:([0-9]+)\:([0-9]+)\s+([0-9]+)\-([0-9]+)\-([0-9]+)\s([0-9]+)\:([0-9]+)\:([0-9]+)\s([0-9]+)/;

		${ipaddr_line}=$1;
		# modify ver1.2.1
		#${year_create_line}=sprintf "%d",$2;
		#${month_create_line}=sprintf "%d",$3;
		#${date_create_line}=sprintf "%d",$4;
		#${hour_create_line}=sprintf "%d",$5;
		#${minute_create_line}=sprintf "%d",$6;
		#${second_create_line}=sprintf "%d",$7;
		#${year_access_line}=sprintf "%d",$8;
		#${month_access_line}=sprintf "%d",$9;
		#${date_access_line}=sprintf "%d",$10;
		#${hour_access_line}=sprintf "%d",$11;
		#${minute_access_line}=sprintf "%d",$12;
		#${second_access_line}=sprintf "%d",$13;
		#${count_comesoon}=sprintf "%d",$14;
		${year_create_tmp}=$2;
		${month_create_tmp}=$3;
		${date_create_tmp}=$4;
		${hour_create_tmp}=$5;
		${minute_create_tmp}=$6;
		${second_create_tmp}=$7;
		${year_access_tmp}=$8;
		${month_access_tmp}=$9;
		${date_access_tmp}=$10;
		${hour_access_tmp}=$11;
		${minute_access_tmp}=$12;
		${second_access_tmp}=$13;
		${count_comesoon_tmp}=$14;
		# modify cancel
		#${year_create_line}=eval { sprintf "%d",$2;}; if ( $@ ) { ${flag_invalidformat} = 1; };
		if ( ( defined ${year_create_tmp} ) and ( ${year_create_tmp} =~ m/^[0-9]+$/ ) ) { ${year_create_line}=${year_create_tmp}; } else { ${year_create_line}="";${flag_invalidformat} = 1;};
		if ( ( defined ${month_create_tmp} ) and ( ${month_create_tmp} =~ m/^[0-9]+$/ ) ) { ${month_create_line}=${month_create_tmp}; } else { ${month_create_line}="";${flag_invalidformat} = 1;};
		if ( ( defined ${date_create_tmp} ) and ( ${date_create_tmp} =~ m/^[0-9]+$/ ) ) { ${date_create_line}=${date_create_tmp}; } else { ${date_create_line}="";${flag_invalidformat} = 1;};
		if ( ( defined ${hour_create_tmp} ) and ( ${hour_create_tmp} =~ m/^[0-9]+$/ ) ) { ${hour_create_line}=${hour_create_tmp}; } else { ${hour_create_line}="";${flag_invalidformat} = 1;};
		if ( ( defined ${minute_create_tmp} ) and ( ${minute_create_tmp} =~ m/^[0-9]+$/ ) ) { ${minute_create_line}=${minute_create_tmp}; } else { ${minute_create_line}="";${flag_invalidformat} = 1;};
		if ( ( defined ${second_create_tmp} ) and ( ${second_create_tmp} =~ m/^[0-9]+$/ ) ) { ${second_create_line}=${second_create_tmp}; } else { ${second_create_line}="";${flag_invalidformat} = 1;};
		if ( ( defined ${year_access_tmp} ) and ( ${year_access_tmp} =~ m/^[0-9]+$/ ) ) { ${year_access_line}=${year_access_tmp}; } else { ${year_access_line}="";${flag_invalidformat} = 1;};
		if ( ( defined ${month_access_tmp}) and ( ${month_access_tmp} =~ m/^[0-9]+$/ ) ) { ${month_access_line}=${month_access_tmp}; } else { ${month_access_line}="";${flag_invalidformat} = 1;};
		if ( ( defined ${date_access_tmp} ) and ( ${date_access_tmp} =~ m/^[0-9]+$/ ) ) { ${date_access_line}=${date_access_tmp}; } else { ${date_access_line}="";${flag_invalidformat} = 1;};
		if ( ( defined ${hour_access_tmp} ) and ( ${hour_access_tmp} =~ m/^[0-9]+$/ ) ) { ${hour_access_line}=${hour_access_tmp}; } else { ${hour_access_line}="";${flag_invalidformat} = 1;};
		if ( ( defined ${minute_access_tmp} ) and ( ${minute_access_tmp} =~ m/^[0-9]+$/ ) ) { ${minute_access_line}=${minute_access_tmp}; } else { ${minute_access_line}="";${flag_invalidformat} = 1;};
		if ( ( defined ${second_access_tmp} ) and ( ${second_access_tmp} =~ m/^[0-9]+$/ ) ) { ${second_access_line}=${second_access_tmp}; } else { ${second_access_line}="";${flag_invalidformat} = 1;};
		if ( ( defined ${count_comesoon_tmp} ) and ( ${count_comesoon_tmp} =~ m/^[0-9]+$/ ) ) { ${count_comesoon}=${count_comesoon_tmp}; } else { ${count_comesoon}=0;};

		# The clear of the backreference
		'' =~ m/(.*)(.*)(.*)(.*)(.*)(.*)(.*)(.*)(.*)(.*)(.*)(.*)(.*)(.*)/;
		
		# add ver1.2.1 if only
		if ( ${flag_invalidformat} == 0 ) {
			if ( ( ${month_create_line} >= 1 ) and
			     ( ${month_access_line} >= 1 ) and
			     ( ${year_create_line} >= 1900) and
			     ( ${year_access_line} >= 1900 ) ){
				${timesirial_create}=timelocal(${second_create_line},${minute_create_line},${hour_create_line},${date_create_line},${month_create_line}-1,${year_create_line}-1900);
				${timesirial_access}=timelocal(${second_access_line},${minute_access_line},${hour_access_line},${date_access_line},${month_access_line}-1,${year_access_line}-1900);
			}
		}
	}

	# return timesiral value     if error undef
	return (${timesirial_create},${timesirial_access},${count_comesoon});
}

# This function erases the information of a host passing 
# the time limit registered with greylist.
# modify ver1.2.3
#sub delete_expirationhostinfo_greylist {
#	my ${greylist}="";
#	my ${count_comesoon}="";
#	my ${timesirial_create};
#	my ${timesirial_access};
#	my ${timestamp_create};
#	my ${timestamp_access};
#	my ${timesirial_now};
#	my ${getline}="";
#	my ${getline_cleaning};
#	my ${objectipaddr};
#
#	( ${greylist} ) = @_;
#
#	# It check whether there is greylist
#	if ( ! -w "${greylist}" ) {
#		write_log LOG_LV_FORCE,LOG_ERR,"-","greylisting: Can't open greylist : ${greylist} ";
#		print "451 temporary problem \r\n";
#		exit;
#	}
#
#	# get timesirial value Of the present
#	${timesirial_now}=time;
#
#	#----------------------------------------------------------------------
#	# It builds a file finished with update in a temporary file.
#
#	# It checks whether other processes do not use atemporary file.
#	while ( -f ${Path_Tmp_Greylist} ) {
#		sleep 1;
#	}
#
#	# It checks whether other processes do not lock greylist.
#	check_greylistlock;
#	# It locks greylist.
#	create_greylistlock;
#
#	# It open greylist read only
#	if ( open(FH_GREY,${greylist}) ) {
#		# It open up temporary file in overwrite mode
#		if ( open(FH_TMP,">${Path_Tmp_Greylist}") ) {
#			# It reads one line from greylist
#			while ( ${getline} = <FH_GREY> ) {
#				# cleaning of reading data
#				${getline_cleaning}=${getline};
#				chomp ${getline_cleaning};
#				${getline_cleaning} =~ s/#.+//;
#				${getline_cleaning} =~ s/^\s+//;
#				${getline_cleaning} =~ s/\s+$//;
#
#				if ( ( defined ${getline_cleaning} ) and
#				     ( ${getline_cleaning} ne ''  ) and
#				     ( ${getline_cleaning} =~ m/^[0-9]+\./ )  ) {
#				     
#					${getline_cleaning} =~ m/^([0-9\.]+)(\s|$)/;
#					${objectipaddr}=$1;
#					
#					( ${timesirial_create} , ${timesirial_access} , ${count_comesoon} ) = get_create_access_time_for_data ${getline_cleaning};
#
#					# add ver1.2.1 if only
#					if ( ( defined ${timesirial_create} ) and ( defined ${timesirial_access} ) ) {
#						if ( ${timesirial_access} != 0 ){
#							if ( ${timesirial_create} == ${timesirial_access} ) {
#								# In the case of a host and it is registered with greylist and is never admitted
#								if ( ${timesirial_create} < ( ${timesirial_now} - ${Sec_Validity_Nogoodhost} ) ) {
#									# It is cancellation of the permission in having had free interval than the last access more than validity
#									${timestamp_create} = conv_logtime ${timesirial_access};
#									write_log LOG_LV_INFO,LOG_INFO,"greylisting","delete ${objectipaddr} (seen once ${timestamp_create})";
#									next ;
#								}
#							}else{
#								# In the case of a host and it is registered with greylist and is  admitting
#								if ( ${timesirial_access} < ( ${timesirial_now} - ${Sec_Validity_Goodhost} ) ) {
#									# It is cancellation of the permission in having had free interval than the last access more than validity
#									${timestamp_create} = conv_logtime ${timesirial_create};
#									${timestamp_access} = conv_logtime ${timesirial_access};
#									write_log LOG_LV_INFO,LOG_INFO,"greylisting","delete ${objectipaddr} (first seen ${timestamp_create}, last ${timestamp_access})";
#									next ;
#								}
#							}
#						}
#					}else{
#						write_log LOG_LV_INFO,LOG_WARN,"greylisting","delete ${objectipaddr} (invalid format)";
#						next;
#					}
#				}
#				# The comment line lines that did not fall under a condition again are
#				# just begun to write
#				printf FH_TMP "%s",${getline};
#			}
#			close FH_TMP;
#		}else{
#			# syslog "err", "%s (%s) ERR - greylisting: Failed to write tmp file : ${Path_Tmp_Greylist} ",${Remotehostip},${Remoteip};
#			write_log LOG_LV_FORCE,LOG_ERR,"greylisting","Failed to write tmp file : ${Path_Tmp_Greylist} ";
#			print "451 temporary problem \r\n";
#			remove_greylistlock;  # add ver1.2.2
#			exit;
#		}
#		
#		close FH_GREY;
#	}else{
#		#syslog "err", "%s (%s) ERR - greylisting: Failed to read greylist : ${greylist} ",${Remotehostip},${Remoteip};
#		write_log LOG_LV_FORCE,LOG_ERR,"greylisting","Failed to read greylist : ${greylist} ";
#		print "451 temporary problem \r\n";
#		
#		remove_greylistlock;  # add ver1.2.2
#		exit;	
#	}
#	
#	# update greylist
#	if ( ${Enable_Flock} == 1 ) {
#		copy_file_flock ${Path_Tmp_Greylist},${greylist};
#	}else{
#		copy_file ${Path_Tmp_Greylist},${greylist};
#	}
#	# unlock greylist
#	remove_greylistlock;
#	
#	# delete temporary file
#	unlink ${Path_Tmp_Greylist};
#}

sub delete_expirationhostinfo_greylist_sub {
	my ${getline};
	my ${getline_cleaning};
	my ${objectipaddr};
	my ${count_comesoon}="";
	my ${timesirial_now};
	my ${timesirial_create};
	my ${timesirial_access};
	my ${timestamp_create};
	my ${timestamp_access};
	
	( ${getline},${timesirial_now} ) = @_;
	
	# cleaning of reading data
	${getline_cleaning}=${getline};
	chomp ${getline_cleaning};
	${getline_cleaning} =~ s/#.+//;
	${getline_cleaning} =~ s/^\s+//;
	${getline_cleaning} =~ s/\s+$//;

	if ( ( defined ${getline_cleaning} ) and
	     ( ${getline_cleaning} ne ''  ) and
	     ( ${getline_cleaning} =~ m/^[0-9]+\./ )  ) {
	     
		${getline_cleaning} =~ m/^([0-9\.]+)(\s|$)/;
		${objectipaddr}=$1;
		
		( ${timesirial_create} , ${timesirial_access} , ${count_comesoon} ) = get_create_access_time_for_data ${getline_cleaning};

		# add ver1.2.1 if only
		if ( ( defined ${timesirial_create} ) and 
		     ( ${timesirial_create} ne '' ) and
		     ( defined ${timesirial_access} ) and
		     ( ${timesirial_access} ne '' ) ) {
			if ( ${timesirial_access} != 0 ){
				if ( ${timesirial_create} == ${timesirial_access} ) {
					# In the case of a host and it is registered with greylist and is never admitted
					if ( ${timesirial_create} < ( ${timesirial_now} - ${Sec_Validity_Nogoodhost} ) ) {
						# It is cancellation of the permission in having had free interval than the last access more than validity
						${timestamp_create} = conv_logtime ${timesirial_access};
						write_log LOG_LV_INFO,LOG_INFO,"greylisting","delete ${objectipaddr} (seen once ${timestamp_create})";
						# modify ver1.2.4
						# return ;
						return undef;
					}
				}else{
					# In the case of a host and it is registered with greylist and is  admitting
					if ( ${timesirial_access} < ( ${timesirial_now} - ${Sec_Validity_Goodhost} ) ) {
						# It is cancellation of the permission in having had free interval than the last access more than validity
						${timestamp_create} = conv_logtime ${timesirial_create};
						${timestamp_access} = conv_logtime ${timesirial_access};
						write_log LOG_LV_INFO,LOG_INFO,"greylisting","delete ${objectipaddr} (first seen ${timestamp_create}, last ${timestamp_access})";
						# modify ver1.2.4
						# return ;
						return undef;
					}
				}
			}
		}else{
			write_log LOG_LV_INFO,LOG_WARN,"greylisting","delete ${objectipaddr} (invalid format)";
			# modify ver1.2.4
			# return ;
			return undef;
		}
	}
	return ${getline};
}

sub delete_expirationhostinfo_greylist_noflock {
	my ${greylist}="";
	my ${timesirial_now};
	my ${getline}="";

	( ${greylist} ) = @_;

	# It check whether there is greylist
	if ( ! -w "${greylist}" ) {
		write_log LOG_LV_FORCE,LOG_ERR,"-","greylisting: Can't open greylist : ${greylist} ";
		print "451 temporary problem \r\n";
		exit;
	}

	# get timesirial value Of the present
	${timesirial_now}=time;

	#----------------------------------------------------------------------
	# It builds a file finished with update in a temporary file.

	# It checks whether other processes do not use atemporary file.
	#while ( -f ${Path_Tmp_Greylist} ) {
	#	sleep 1;
	#}
	check_greylisttmp;

	# It checks whether other processes do not lock greylist.
	check_greylistlock;
	# It locks greylist.
	create_greylistlock;

	# It open greylist read only
	if ( open(FH_GREY,${greylist}) ) {
		# It open up temporary file in overwrite mode
		if ( open(FH_TMP,">${Path_Tmp_Greylist}") ) {
			# It reads one line from greylist
			while ( ${getline} = <FH_GREY> ) {
			
				if ( delete_expirationhostinfo_greylist_sub ${getline},${timesirial_now} ) {
					# The comment line lines that did not fall under a condition again are
					# just begun to write
					printf FH_TMP "%s",${getline};
				}
			}
			close FH_TMP;
		}else{
			# syslog "err", "%s (%s) ERR - greylisting: Failed to write tmp file : ${Path_Tmp_Greylist} ",${Remotehostip},${Remoteip};
			write_log LOG_LV_FORCE,LOG_ERR,"greylisting","Failed to write tmp file : ${Path_Tmp_Greylist} ";
			print "451 temporary problem \r\n";
			remove_greylistlock;  # add ver1.2.2
			remove_greylisttmp;   # add ver1.2.3
			exit;
		}
		
		close FH_GREY;
	}else{
		#syslog "err", "%s (%s) ERR - greylisting: Failed to read greylist : ${greylist} ",${Remotehostip},${Remoteip};
		write_log LOG_LV_FORCE,LOG_ERR,"greylisting","Failed to read greylist : ${greylist} ";
		print "451 temporary problem \r\n";
		
		remove_greylistlock;  # add ver1.2.2
		remove_greylisttmp;   # add ver1.2.3
		exit;	
	}
	
	# update greylist
	copy_file ${Path_Tmp_Greylist},${greylist};

	# unlock greylist
	remove_greylistlock;
	
	# delete temporary file
	unlink ${Path_Tmp_Greylist};
}

sub delete_expirationhostinfo_greylist_flock {
	my ${greylist}="";
	my ${timesirial_now};
	my ${getline}="";

	( ${greylist} ) = @_;

	# It check whether there is greylist
	if ( ! -w "${greylist}" ) {
		write_log LOG_LV_FORCE,LOG_ERR,"-","greylisting: Can't open greylist : ${greylist} ";
		print "451 temporary problem \r\n";
		exit;
	}

	# get timesirial value Of the present
	${timesirial_now}=time;

	#----------------------------------------------------------------------
	# It builds a file finished with update in a temporary file.
	
	# set alarm
	$SIG{'ALRM'} = "sigexit_check_greylistlock_timeout";
	alarm ${Sec_Smtptimeout};

	# It open greylist
	if ( sysopen(FH_GREY,${greylist},O_RDWR|O_CREAT|O_SYNC) ) {
		flock FH_GREY,LOCK_EX;
		# It open up temporary file in overwrite mode
		if ( sysopen(FH_TMP,${Path_Tmp_Greylist},O_RDWR|O_CREAT|O_SYNC) ) {
			flock FH_TMP,LOCK_EX;
			
			seek(FH_GREY,0,0);
			
			truncate( FH_TMP,0 );
			seek(FH_TMP,0,0);
			
			# It reads one line from greylist
			while ( ${getline} = <FH_GREY> ) {
				if ( delete_expirationhostinfo_greylist_sub ${getline},${timesirial_now} ) {
					# The comment line lines that did not fall under a condition again are
					# just begun to write
					printf FH_TMP "%s",${getline};
				}
			}
			
			# update greylist
			if ( ! copy_file_FH( *FH_TMP, *FH_GREY ) ) {
				write_log LOG_LV_FORCE,LOG_ERR,"greylisting","Failed to write tmp file : ${Path_Tmp_Greylist} ";
				print "451 temporary problem \r\n";
				remove_greylistlock;  # add ver1.2.2
				remove_greylisttmp;   # add ver1.2.3
				exit;
			}
			#truncate( FH_GREY,0 );
			#seek(FH_GREY,0,0);
			#seek(FH_TMP,0,0);
			#while ( ${getline} = <FH_TMP> ){
			#	printf FH_GREY "%s",${getline};
			#}
			close FH_TMP;
			# delete temporary file
			unlink ${Path_Tmp_Greylist};
		}else{
			# syslog "err", "%s (%s) ERR - greylisting: Failed to write tmp file : ${Path_Tmp_Greylist} ",${Remotehostip},${Remoteip};
			write_log LOG_LV_FORCE,LOG_ERR,"greylisting","Failed to write tmp file : ${Path_Tmp_Greylist} ";
			print "451 temporary problem \r\n";
			remove_greylistlock;  # add ver1.2.2
			remove_greylisttmp;   # add ver1.2.3
			exit;
		}
		
		close FH_GREY;
	}else{
		#syslog "err", "%s (%s) ERR - greylisting: Failed to read greylist : ${greylist} ",${Remotehostip},${Remoteip};
		write_log LOG_LV_FORCE,LOG_ERR,"greylisting","Failed to read greylist : ${greylist} ";
		print "451 temporary problem \r\n";
		
		remove_greylistlock;  # add ver1.2.2
		remove_greylisttmp;   # add ver1.2.3
		exit;	
	}
	
	# unset alarm
	alarm 0;
	$SIG{'ALRM'} = "";
	
}

sub delete_expirationhostinfo_greylist {
	my ${greylist}="";

	( ${greylist} ) = @_;
	
	if ( ${Enable_Flock} == 1 ) {
		delete_expirationhostinfo_greylist_flock ${greylist};
	}else{
		delete_expirationhostinfo_greylist_noflock ${greylist}
	}
}


# This function erases the information of a host registered with greylist
sub delete_hostinfo_greylist_noflock {
	my ${greylist}="";
	my ${ipaddr}="";
	my ${getline}="";

	( ${greylist} , ${ipaddr}) = @_;


	# check whether there is greylist
	if ( ! -w "${greylist}" ) {
		write_log LOG_LV_FORCE,LOG_ERR,"greylisting","Can't open greylist : ${greylist} ";
		print "451 temporary problem \r\n";
		exit;
	}

	#----------------------------------------------------------------------
	# build a file finished with update in a temporary file.

	# check whether other processes do not use atemporary file.
	#while ( -f ${Path_Tmp_Greylist} ) {
	#	sleep 1;
	#}
	check_greylisttmp;

	# check whether other processes do not lock greylist.
	check_greylistlock;
	# lock greylist.
	create_greylistlock;

	# open greylist read only
	if ( open(FH_GREY,${greylist}) ) {
		# open up temporary file in overwrite mode
		if ( open(FH_TMP,">${Path_Tmp_Greylist}") ) {
			# read one line from greylist
			while ( ${getline} = <FH_GREY> ) {

				# check it whether it is a deletion object record
				if ( ${getline} =~ m/^${ipaddr}(\s.+|$)/ ) {
					# If it is an object,  do not write in it
					write_log LOG_LV_DEBUG,LOG_INFO,"greylisting","delete ${ipaddr}";
				}else{
					# If there is not it in an object, write the line data 
					printf FH_TMP "%s",${getline};
				}
			}
			close FH_TMP;
		}else{
			write_log LOG_LV_FORCE,LOG_ERR,"greylisting","Failed to write tmp file : ${Path_Tmp_Greylist} ";
			print "451 temporary problem \r\n";
			remove_greylistlock;  # add ver1.2.2
			remove_greylisttmp;   # add ver1.2.3
			exit;
		}
		close FH_GREY;
	}else{
		write_log LOG_LV_FORCE,LOG_ERR,"greylisting","Failed to read greylist : ${greylist} ";
		print "451 temporary problem \r\n";
		remove_greylistlock;  # add ver1.2.2
		remove_greylisttmp;   # add ver1.2.3
		exit;	
	}
	
	# update greylist
	copy_file ${Path_Tmp_Greylist},${greylist};
	
	# unlock greylist
	remove_greylistlock;

	# delete temporary file
	unlink ${Path_Tmp_Greylist};
}
# add ver1.2.3
sub delete_hostinfo_greylist_flock {
	my ${greylist}="";
	my ${ipaddr}="";
	my ${getline}="";

	( ${greylist} , ${ipaddr}) = @_;


	# check whether there is greylist
	if ( ! -w "${greylist}" ) {
		write_log LOG_LV_FORCE,LOG_ERR,"greylisting","Can't open greylist : ${greylist} ";
		print "451 temporary problem \r\n";
		exit;
	}

	#----------------------------------------------------------------------
	# build a file finished with update in a temporary file.

	# set alarm
	$SIG{'ALRM'} = "sigexit_check_greylistlock_timeout";
	alarm ${Sec_Smtptimeout};
	
	# open greylist read only
	if ( sysopen(FH_GREY,${greylist},O_RDWR|O_CREAT|O_SYNC) ) {
		flock FH_GREY,LOCK_EX;
		# open up temporary file in overwrite mode
		if ( sysopen(FH_TMP,${Path_Tmp_Greylist},O_RDWR|O_CREAT|O_SYNC) ) {
			flock FH_TMP,LOCK_EX;

			seek(FH_GREY,0,0);
			
			# clear temporary file
			truncate( FH_TMP,0 );
			seek(FH_TMP,0,0);
			
			# read one line from greylist
			while ( ${getline} = <FH_GREY> ) {

				# check it whether it is a deletion object record
				if ( ${getline} =~ m/^${ipaddr}(\s.+|$)/ ) {
					# If it is an object,  do not write in it
					write_log LOG_LV_DEBUG,LOG_INFO,"greylisting","delete ${ipaddr}";
				}else{
					# If there is not it in an object, write the line data 
					printf FH_TMP "%s",${getline};
				}
			}
			
			# update greylist
			if ( ! copy_file_FH( *FH_TMP, *FH_GREY ) ) {
				write_log LOG_LV_FORCE,LOG_ERR,"greylisting","Failed to write tmp file : ${Path_Tmp_Greylist} ";
				print "451 temporary problem \r\n";
				remove_greylistlock;  # add ver1.2.2
				remove_greylisttmp;   # add ver1.2.3
				exit;
			}
			#truncate( FH_GREY,0 );
			#seek(FH_GREY,0,0);
			#seek(FH_TMP,0,0);
			#while ( ${getline} = <FH_TMP> ) {
			#	printf FH_GREY "%s",${getline};
			#}
			close FH_TMP;
			# delete temporary file
			unlink ${Path_Tmp_Greylist};
		}else{
			write_log LOG_LV_FORCE,LOG_ERR,"greylisting","Failed to write tmp file : ${Path_Tmp_Greylist} ";
			print "451 temporary problem \r\n";
			remove_greylistlock;  # add ver1.2.2
			remove_greylisttmp;   # add ver1.2.3
			exit;
		}
		close FH_GREY;
	}else{
		write_log LOG_LV_FORCE,LOG_ERR,"greylisting","Failed to read greylist : ${greylist} ";
		print "451 temporary problem \r\n";
		remove_greylistlock;  # add ver1.2.2
		remove_greylisttmp;   # add ver1.2.3
		exit;	
	}
	
	# unset alarm
	alarm 0;
	$SIG{'ALRM'} = "";
	
}

sub delete_hostinfo_greylist {
	my ${greylist}="";
	my ${ipaddr}="";

	( ${greylist} , ${ipaddr}) = @_;
	
	if ( ${Enable_Flock} == 1 ) {
		delete_hostinfo_greylist_flock ${greylist},${ipaddr};
	}else{
		delete_hostinfo_greylist_noflock ${greylist},${ipaddr};
	}
}


# This function registers the information of the host with greylist
sub set_hostinfo_greylist_noflock {
	my ${greylist}="";
	my ${ipaddr}="";
	my ${timesirial_create}="";
	my ${timesirial_access}="";
	my ${count_comesoon}="";
	my ${second_create}="";
	my ${minute_create}="";
	my ${hour_create}="";
	my ${date_create}="";
	my ${month_create}="";
	my ${year_create}="";
	my ${second_access}="";
	my ${minute_access}="";
	my ${hour_access}="";
	my ${date_access}="";
	my ${month_access}="";
	my ${year_access}="";
	my ${getline}="";
	my ${flg_update};


	( ${greylist} , ${ipaddr},${timesirial_create},${timesirial_access},${count_comesoon}) = @_;
	
	# convert timesirial value into the value that a person can understand
	(${second_create},${minute_create},${hour_create},${date_create},${month_create},${year_create})=localtime(${timesirial_create});
	(${second_access},${minute_access},${hour_access},${date_access},${month_access},${year_access})=localtime(${timesirial_access});
	${month_create}=${month_create}+1;
	${year_create}=${year_create}+1900;
	${month_access}=${month_access}+1;
	${year_access}=${year_access}+1900;
	
	# modify ver1.2.4
	# if ( ! defined ${count_comesoon} ){	
	if ( ( ! defined ${count_comesoon} ) or ( ${count_comesoon} eq '' )  ){
		${count_comesoon} = 0;
	}


	# check whether there is greylist
	if ( ! -w "${greylist}" ) {
		write_log LOG_LV_FORCE,LOG_ERR,"greylisting","Can't open greylist : ${greylist} ";
		print "451 temporary problem \r\n";
		exit;
	}
	#----------------------------------------------------------------------
	# build a file finished with update in a temporary file.

	# check whether other processes do not use atemporary file.
	#while ( -f ${Path_Tmp_Greylist} ) {
	#	sleep 1;
	#}
	check_greylisttmp;

	# check whether other processes do not lock greylist.
	check_greylistlock;
	# lock greylist.
	create_greylistlock;

	# open greylist read only
	if ( open(FH_GREY,${greylist}) ) {
		# open up temporary file in overwrite mode
		if ( open(FH_TMP,">${Path_Tmp_Greylist}") ) {
			${flg_update}=0;
			# read one line from greylist
			while ( ${getline} = <FH_GREY> ) {

				# check it whether it is an update object record
				if ( ${getline} =~ m/^${ipaddr}(\s+|$)/ ) {
					# If it is an object,  update it
					printf FH_TMP   "%-15s %04d-%02d-%02d %02d:%02d:%02d %04d-%02d-%02d %02d:%02d:%02d %d\n",${ipaddr},${year_create},${month_create},${date_create},${hour_create},${minute_create},${second_create},${year_access},${month_access},${date_access},${hour_access},${minute_access},${second_access},${count_comesoon};
					${flg_update}=1;
				}else{
					# If there is not it in an object, write the line data 
					printf FH_TMP "%s",${getline};
				}
			}
			if ( ${flg_update} == 0 ){
				# When there was not an update object record,  adds in the last of the file
				printf FH_TMP   "%-15s %04d-%02d-%02d %02d:%02d:%02d %04d-%02d-%02d %02d:%02d:%02d %d\n",${ipaddr},${year_create},${month_create},${date_create},${hour_create},${minute_create},${second_create},${year_access},${month_access},${date_access},${hour_access},${minute_access},${second_access},${count_comesoon};
			}
			close FH_TMP;
		}else{
			write_log LOG_LV_FORCE,LOG_ERR,"greylisting","Failed to write tmp file : ${Path_Tmp_Greylist} ";
			print "451 temporary problem \r\n";
			remove_greylistlock;  # add ver1.2.2
			remove_greylisttmp;   # add ver1.2.3
			exit;
		}
		close FH_GREY;
	}else{
		write_log LOG_LV_FORCE,LOG_ERR,"greylisting","Failed to read greylist : ${greylist} ";
		print "451 temporary problem \r\n";
		remove_greylistlock;  # add ver1.2.2
		remove_greylisttmp;   # add ver1.2.3
		exit;	
	}

	# update greylist
	copy_file ${Path_Tmp_Greylist},${greylist};
	#  unlock greylist
	remove_greylistlock;

	# delete temporary file
	unlink ${Path_Tmp_Greylist};

}
# add ver1.2.3 
sub set_hostinfo_greylist_flock {
	my ${greylist}="";
	my ${ipaddr}="";
	my ${timesirial_create}="";
	my ${timesirial_access}="";
	my ${count_comesoon}="";
	my ${second_create}="";
	my ${minute_create}="";
	my ${hour_create}="";
	my ${date_create}="";
	my ${month_create}="";
	my ${year_create}="";
	my ${second_access}="";
	my ${minute_access}="";
	my ${hour_access}="";
	my ${date_access}="";
	my ${month_access}="";
	my ${year_access}="";
	my ${getline}="";
	my ${flg_update};


	( ${greylist} , ${ipaddr},${timesirial_create},${timesirial_access},${count_comesoon}) = @_;
	
	# convert timesirial value into the value that a person can understand
	(${second_create},${minute_create},${hour_create},${date_create},${month_create},${year_create})=localtime(${timesirial_create});
	(${second_access},${minute_access},${hour_access},${date_access},${month_access},${year_access})=localtime(${timesirial_access});
	${month_create}=${month_create}+1;
	${year_create}=${year_create}+1900;
	${month_access}=${month_access}+1;
	${year_access}=${year_access}+1900;
	
	
	# modify ver1.2.4
	# if ( ! defined ${count_comesoon} ) {
	if ( ( ! defined ${count_comesoon} ) or ( ${count_comesoon} eq '' )  ){
		${count_comesoon} = 0;
	}


	# check whether there is greylist
	if ( ! -w "${greylist}" ) {
		write_log LOG_LV_FORCE,LOG_ERR,"greylisting","Can't open greylist : ${greylist} ";
		print "451 temporary problem \r\n";
		exit;
	}
	#----------------------------------------------------------------------
	# build a file finished with update in a temporary file.

	# set alarm
	$SIG{'ALRM'} = "sigexit_check_greylistlock_timeout";
	alarm ${Sec_Smtptimeout};
	
	# open greylist read only
	if ( sysopen(FH_GREY,${greylist},O_RDWR|O_CREAT|O_SYNC) ) {
		flock FH_GREY,LOCK_EX;
		# open up temporary file in overwrite mode
		if ( sysopen(FH_TMP,${Path_Tmp_Greylist},O_RDWR|O_CREAT|O_SYNC) ) {
			flock FH_TMP,LOCK_EX;
			
			seek(FH_GREY,0,0);
			
			truncate( FH_TMP,0 );
			seek(FH_TMP,0,0);
			
			${flg_update}=0;
			# read one line from greylist
			while ( ${getline} = <FH_GREY> ) {

				# check it whether it is an update object record
				if ( ${getline} =~ m/^${ipaddr}(\s+|$)/ ) {
					# If it is an object,  update it
					printf FH_TMP   "%-15s %04d-%02d-%02d %02d:%02d:%02d %04d-%02d-%02d %02d:%02d:%02d %d\n",${ipaddr},${year_create},${month_create},${date_create},${hour_create},${minute_create},${second_create},${year_access},${month_access},${date_access},${hour_access},${minute_access},${second_access},${count_comesoon};
					${flg_update}=1;
				}else{
					# If there is not it in an object, write the line data 
					printf FH_TMP "%s",${getline};
				}
			}
			if ( ${flg_update} == 0 ){
				# When there was not an update object record,  adds in the last of the file
				printf FH_TMP   "%-15s %04d-%02d-%02d %02d:%02d:%02d %04d-%02d-%02d %02d:%02d:%02d %d\n",${ipaddr},${year_create},${month_create},${date_create},${hour_create},${minute_create},${second_create},${year_access},${month_access},${date_access},${hour_access},${minute_access},${second_access},${count_comesoon};
			}
			
			# update greylist
			if ( ! copy_file_FH( *FH_TMP, *FH_GREY ) ) {
				write_log LOG_LV_FORCE,LOG_ERR,"greylisting","Failed to write tmp file : ${Path_Tmp_Greylist} ";
				print "451 temporary problem \r\n";
				remove_greylistlock;  # add ver1.2.2
				exit;
			}
			#truncate( FH_GREY,0 );
			#seek(FH_GREY,0,0);
			#
			#seek(FH_TMP,0,0);
			#
			#while ( ${getline} = <FH_TMP> ){
			#	printf FH_GREY "%s",${getline};
			#}
			close FH_TMP;
			# delete temporary file
			unlink ${Path_Tmp_Greylist};
		}else{
			write_log LOG_LV_FORCE,LOG_ERR,"greylisting","Failed to write tmp file : ${Path_Tmp_Greylist} ";
			print "451 temporary problem \r\n";
			remove_greylistlock;  # add ver1.2.2
			exit;
		}
		close FH_GREY;
	}else{
		write_log LOG_LV_FORCE,LOG_ERR,"greylisting","Failed to read greylist : ${greylist} ";
		print "451 temporary problem \r\n";
		remove_greylistlock;  # add ver1.2.2
		remove_greylisttmp;   # add ver1.2.3
		exit;	
	}
	
	# unset alarm
	alarm 0;
	$SIG{'ALRM'} = "";
}

sub set_hostinfo_greylist {
	my ${greylist};
	my ${ipaddr};
	my ${timesirial_create};
	my ${timesirial_access};
	my ${count_comesoon};
	
	( ${greylist} , ${ipaddr},${timesirial_create},${timesirial_access},${count_comesoon}) = @_;
	
	if ( ${Enable_Flock} == 1 ) {
		set_hostinfo_greylist_flock ${greylist} , ${ipaddr},${timesirial_create},${timesirial_access},${count_comesoon};
	}else{
		set_hostinfo_greylist_noflock ${greylist} , ${ipaddr},${timesirial_create},${timesirial_access},${count_comesoon};	
	}
}

# format of greylist
#   ipaddr create_date create_time access_date access_time come_soon_count
#   xxx.xxx.xxx.xxxx YYYY-MM-DD hh:mm:ss YYYY-MM-DD hh:mm:ss [0-9]+

# This function gets information on the host registered in greylist.

sub get_hostinfo_sub {
	my ${getline};
	my ${ipaddr};
	
	( ${getline},${ipaddr} ) = @_;
	
	# cleaning of reading data
	chomp ${getline};
	${getline} =~ s/#.+//;
	${getline} =~ s/^\s+//;
	${getline} =~ s/\s+$//;

	if (  ${getline} eq '' ){
		# modify ver1.2.4
		# return ;
		return undef ;
	}

	if ( ${getline} =~ m/^${ipaddr}(\s+|$)/ ) {
		return ${getline};
	}

}

sub get_hostinfo_noflock {
	my ${greylist}="";
	my ${ipaddr}="";
	my ${hostinfo}="";
	my ${getline}="";

	( ${greylist} , ${ipaddr} ) = @_;

	# check whether there is greylist
	if ( ! -f "${greylist}" ) {
		write_log LOG_LV_FORCE,LOG_ERR,"greylisting","Can't open greylist : ${greylist}";
		print "451 temporary problem \r\n";
		exit;
	}

	# check whether other processes do not lock greylist.
	check_greylistlock;

	#open greylist read only
	if ( open(FH_GREY,${greylist}) ) {
		while ( ${getline} = <FH_GREY> ) {
#			# cleaning of reading data
#			chomp ${getline};
#			${getline} =~ s/#.+//;
#			${getline} =~ s/^\s+//;
#			${getline} =~ s/\s+$//;
#
#			if (  ${getline} eq '' ){
#				next;
#			}
#
#			if ( ${getline} =~ m/^${ipaddr}(\s+|$)/ ) {
#				${hostinfo} = ${getline};
#				last;
#			}
#			
			${hostinfo} = get_hostinfo_sub ${getline},${ipaddr};
			# modify ver1.2.4
			# if ( defined ${hostinfo} ){
			if ( ( defined ${hostinfo} ) and ( ${hostinfo} ne '' ) ){
				last;
			}
		}
		close FH_GREY;
	}else{
		write_log LOG_LV_FORCE,LOG_ERR,"greylisting","Failed to read greylist : ${greylist} ";
		print "451 temporary problem \r\n";
		exit;	
	}

	# add ver1.2.5
	if ( ( defined ${hostinfo} ) and ( ${hostinfo} ne '' ) ) {
		# return the whole line data of the object address(finished with cleaning)
		return ${hostinfo};
	}else{
		return undef;
	}
}
sub get_hostinfo_flock {
	my ${greylist}="";
	my ${ipaddr}="";
	my ${hostinfo}="";
	my ${getline}="";

	( ${greylist} , ${ipaddr} ) = @_;

	# check whether there is greylist
	if ( ! -f "${greylist}" ) {
		write_log LOG_LV_FORCE,LOG_ERR,"greylisting","Can't open greylist : ${greylist}";
		print "451 temporary problem \r\n";
		exit;
	}

	# set alarm
	$SIG{'ALRM'} = "sigexit_check_greylistlock_timeout";
	alarm ${Sec_Smtptimeout};
	
	#open greylist read only
	if ( sysopen(FH_GREY,${greylist},O_RDONLY) ) {
		flock FH_GREY,LOCK_SH;
		
		while ( ${getline} = <FH_GREY> ) {
			${hostinfo} = get_hostinfo_sub ${getline},${ipaddr};
			# modify ver1.2.4
			# if ( defined ${hostinfo} ){
			if ( ( defined ${hostinfo} ) and ( ${hostinfo} ne '' ) ){
				last;
			}
		}
		close FH_GREY;
	}else{
		write_log LOG_LV_FORCE,LOG_ERR,"greylisting","Failed to read greylist : ${greylist} ";
		print "451 temporary problem \r\n";
		exit;	
	}
	
	# unset alarm
	alarm 0;
	$SIG{'ALRM'} = "";

	# add ver1.2.5
	if ( ( defined ${hostinfo} ) and ( ${hostinfo} ne '' ) ) {
		# return the whole line data of the object address(finished with cleaning)
		return ${hostinfo};
	}else{
		return undef;
	}
}

sub get_hostinfo {
	my ${greylist}="";
	my ${ipaddr}="";
	my ${hostinfo}="";

	( ${greylist} , ${ipaddr} ) = @_;
	
	if ( ${Enable_Flock} == 1 ) {
		${hostinfo} = get_hostinfo_flock ${greylist} , ${ipaddr};
	}else{
		${hostinfo} = get_hostinfo_noflock ${greylist} , ${ipaddr}
	}
	
	return ${hostinfo};
}

# This function acquires various time for registered host information from greylist
sub get_create_access_time_greylist {
	my ${greylist};
	my ${ipaddr};
	my ${hostinfo}="";

	my ${count_comesoon}="";
	my ${timesirial_create};
	my ${timesirial_access};


	( ${greylist} , ${ipaddr} ) = @_;

	# get the data of the object address
	${hostinfo} = get_hostinfo ${greylist},${ipaddr};
	# modify ver1.2.4
	# if ( defined ${hostinfo} )  {
	if ( ( defined ${hostinfo} ) and (  ${hostinfo} ne '' ) ) {
		(${timesirial_create},${timesirial_access},${count_comesoon}) = get_create_access_time_for_data "${hostinfo}";
	}

	# return timesirial value
	return (${timesirial_create},${timesirial_access},${count_comesoon});
}

# This function updates information in time for host information registered with greylist
sub set_access_time_now_greylist {
	my ${greylist}="";
	my ${ipaddr}="";
	my ${count_comesoon}="";
	my ${timesirial_create};
	my ${timesirial_access};
	my ${timesirial_now};

	( ${greylist} , ${ipaddr} ) = @_;

	(${timesirial_create},${timesirial_access},${count_comesoon}) = get_create_access_time_greylist ${greylist},${ipaddr};

	${timesirial_now} = time ;
	${timesirial_access} = ${timesirial_now};
	if ( ( ! defined ${timesirial_create} ) or ( ${timesirial_create} eq '0' ) or ( ${timesirial_create} eq '' ) ) {
		${timesirial_create} = ${timesirial_now};
	}
	set_hostinfo_greylist ${greylist},${ipaddr},${timesirial_create},${timesirial_access},${count_comesoon};
}

# The function of this function puts off the corrected time of the file 
# in the past only in connection suppression time and registers myself. 
# By it this; can be connected without refusing it 
# even if there was connection from a host just after that.
sub set_create_access_time_soonaccept_greylist {
	my ${greylist}="";
	my ${ipaddr}="";
	my ${count_comesoon}="";
	my ${timesirial_create};
	my ${timesirial_access};
	my ${timesirial_now};
	my ${timesirial_create_modify};
	
	( ${greylist} , ${ipaddr} ) = @_;

	(${timesirial_create},${timesirial_access},${count_comesoon}) = get_create_access_time_greylist ${greylist},${ipaddr};
	
	# add ver1.2.1 if only
	if ( defined ${timesirial_create} and ${timesirial_create} ne '' ) {
		${timesirial_now} = time ;
		
		if ( ${timesirial_create} > ( ${timesirial_now} - ${Sec_Suppression_Connect} ) ) {
			${timesirial_create_modify} = ${timesirial_create} - ${Sec_Suppression_Connect};
			
			set_hostinfo_greylist ${greylist},${ipaddr},${timesirial_create_modify},${timesirial_access},${count_comesoon};
		}
	}
}

# This function updates "comesoon count" of host information registered with greylist
sub countup_comesooncount_greylist {
	my ${greylist}="";
	my ${ipaddr}="";
	my ${count_comesoon}="";
	my ${timesirial_create};
	my ${timesirial_access};
	my ${timesirial_now};

	( ${greylist} , ${ipaddr} ) = @_;

	(${timesirial_create},${timesirial_access},${count_comesoon}) = get_create_access_time_greylist ${greylist},${ipaddr};
	
	# add ver1.2.1 if only
	if ( defined ${timesirial_create} and ${timesirial_create} ne '' ) {
		${count_comesoon}++;
		set_hostinfo_greylist ${greylist},${ipaddr},${timesirial_create},${timesirial_access},${count_comesoon};
	}
}

sub check_lastdeletegreylist {
	my ${deletetime};
	
	if ( -f ${Path_Lastdeletetime_Greylist} ) {
		# get deletion processing time of the last time
		${deletetime} = (stat ${Path_Lastdeletetime_Greylist})[9];

		if ( ${deletetime} < ( time - ${Sec_Checkinterval_Greylist} ) ) {
			return 1;
		}
	}else{
		create_file ${Path_Lastdeletetime_Greylist};
	}
	return 0;
}

sub delete_expirationhostinfo_greylist_main {
	if ( check_lastdeletegreylist == 1 ) {
		delete_expirationhostinfo_greylist ${Path_Greylist};
		# add ver1.2.1
		create_file ${Path_Lastdeletetime_Greylist};
	}
}


# -------------------------------------

#sub get_ipfilename {
sub get_objectipaddr {
	my ${ipaddr} ;

	( ${ipaddr} ) = @_;

	if ( ${Mode_Ipfile_Thirdoctet} == 1 ) {
		${ipaddr} =~ /^(\d+\.\d+\.\d+\.)\d+$/;
		return $1;
	}else {
		return ${ipaddr};
	}
}

sub exec_next {
	if ( ! exec @ARGV ) {
		print "450 temporary problem - failed to start next process\r\n";
		
		write_log LOG_LV_FORCE,LOG_ERR,"-","failed to exec next : @ARGV";
		exit;
	}
}

sub load_patternfile {
	my @list;
	my ${getline};
	my ${getline_regexp};
	my ${cmpregexp};

	my ( ${patternfile} ) = @_ ;
	if ( open(FH_PTNFL,${patternfile}) ) {
		while( ${getline} = <FH_PTNFL>) {
			chomp ${getline};
			${getline} =~ s/#.*$//;
			${getline} =~ s/^\s+//;
			${getline} =~ s/\s+$//;
			${getline} =~ /^(\S+)\s*(.*)$/;
			if ( ( ! defined ${getline} ) or ( ${getline} eq '' ) ){
				next;
			}
			${getline_regexp} = "";
			${getline_regexp} = $1;

			if ( ${getline_regexp} eq '' ) {
				next;
			}
			${getline_regexp} =~ s@^/@@;
			${getline_regexp} =~ s@/$@@;

			if ( ( ! defined ${getline_regexp} ) or 
			     ( ${getline_regexp} eq '' ) or
			     ( ${getline_regexp} =~ /^\s+$/ ) )
			{
				next ;
			}
			undef ${cmpregexp};
			
			# add ver1.3.1
			if ( ${Enable_Eval} == 1 ) {
				eval { ${cmpregexp} = qr/${getline_regexp}/i;};
			}else{
				${cmpregexp} = qr/${getline_regexp}/i;
			}
			if ( defined ${cmpregexp} and ${cmpregexp} ne '' ) {
				push @list , ${cmpregexp};
			} else {
				write_log LOG_LV_FORCE,LOG_WARN,"-","${patternfile} of line $. regular expression error";
			}
		}
		close FH_PTNFL;
	}else{
		write_log LOG_LV_FORCE,LOG_ERR,"-","Failed to open file : ${patternfile}";
	}
	return @list;
}

# It checks whether host name is included in a list file
sub check_hostname {
	my ( ${listfile},${remotehostname} ) = @_;

	my @list = load_patternfile(${listfile});

	foreach my ${reg} ( @list ) {
		if ( ${remotehostname} =~ /${reg}/ ) {
			return ${remotehostname};
		}
	}
	# modify ver1.2.4
	# return ;
	return undef;
}

# ver1.1 add
sub check_helohostname {
	my ( ${listfile},${remotehostname},${remotehostip} ) = @_;

	my ${return_hostip}="";
	my ${getline};
	my ${getline_regexphelo};
	my ${getline_regexphostip};
	my ${cmpregexphelo};
	my ${cmpregexphostip};

	if ( open(FH_PTNFL,${listfile}) ) {
		${return_hostip}="";
		while( ${getline} = <FH_PTNFL>) {
			chomp ${getline};
			${getline} =~ s/#.*$//;  # remove comment
			${getline} =~ s/^\s+//;  # remove start space
			${getline} =~ s/\s+$//;  # remove end space
			${getline} =~ /^(\S+)\s*(.*)$/;
			if ( ( ! defined ${getline} ) or ( ${getline} eq '' ) ){
				next;
			}
			
			# get helohost and ipaddr
			${getline_regexphelo} = "";
			${getline_regexphelo} = $1;
			${getline_regexphostip} = "";
			${getline_regexphostip} = $2;

			if ( ${getline_regexphelo} eq '' ) {
				next;
			}

			${getline_regexphelo} =~ s@^/@@;   # remove start /
			${getline_regexphelo} =~ s@/$@@;   # remove end /

			if ( ( ! defined ${getline_regexphelo} ) or
				( ${getline_regexphelo} eq '' ) or
				( ${getline_regexphelo} =~ /^\s+$/ ) )
			{
				next ;
			}
			
			undef ${cmpregexphelo};    # reset
			${cmpregexphelo} = qr/${getline_regexphelo}/i;  # compile regexp
			if ( ! ( defined ${cmpregexphelo} and ${cmpregexphelo} ne '' )  ) {
				write_log LOG_LV_FORCE,LOG_WARN,"-","${listfile} of line $. regular expression error";
			}
			
			
			${getline_regexphostip} =~ s@^/@@;   # remove start /
			${getline_regexphostip} =~ s@/$@@;   # remove end /
			undef ${cmpregexphostip};
			if ( ( defined ${getline_regexphostip} ) and ( ${getline_regexphostip} ne '' ) ) {
				${cmpregexphostip} = qr/${getline_regexphostip}/i;
				if ( ! ( defined ${cmpregexphostip} and ${cmpregexphostip} ne '' ) ){
					write_log LOG_LV_FORCE,LOG_WARN,"-","${listfile} of line $. regular expression error";
				}
			}
			
			
			# ---- check helohost ----
			if ( ${remotehostname} =~ /${cmpregexphelo}/ ) {
				if ( ! ( defined ${cmpregexphostip} and ${cmpregexphostip} ne '' ) ) {
					${return_hostip}=${remotehostname};
					last;
				}else{
					if ( ${remotehostip} =~ /${cmpregexphostip}/ ) {
						${return_hostip}=${remotehostname};
						last;
					}
				}
			}
		}
		close FH_PTNFL;
		if ( defined ${return_hostip} and ${return_hostip} ne '' ) {
			return ${return_hostip};
		}
	}else{
		write_log LOG_LV_FORCE,LOG_ERR,"-","Failed to open file : ${listfile}";
	}
}

sub sigexit_tarpit {
	write_log LOG_LV_INFO,LOG_NG,"tarpit","can't wait";
	exit;
}

sub tarpit {

	my ( ${Remoteip} ) = @_;

	$SIG{'INT'} = $SIG{'TERM'} = $SIG{'HUP'} = $SIG{'PIPE'} = $SIG{'QUIT'} = "sigexit_tarpit";

	write_log LOG_LV_INFO,LOG_INFO,"tarpit","start";
	sleep ${Sec_Tarpit};

	$SIG{'INT'} = $SIG{'TERM'} = $SIG{'HUP'} = $SIG{'PIPE'} = $SIG{'QUIT'} = "";

	if ( ${Mode_Accept_Tarpitwaithost} == 1 ) {
		write_log LOG_LV_INFO,LOG_OK,"tarpit","accept waited ";
		exec_next;
	}
	write_log LOG_LV_INFO,LOG_INFO,"tarpit","end";
}

sub sigexit_smtpdummy_timeout {
	write_log LOG_LV_INFO,LOG_NG,"SMTP","connection closed (timeout)";
	
	# add ver1.2
	if ( ${Mode_SmtpTimeout_Deny} == 1 ) {
		if ( get_hostinfo ${Path_Greylist},${Objectipaddr} ){
			delete_hostinfo_greylist ${Path_Greylist},${Objectipaddr};
		}
	}
	
	print "421 ${Domain} timeout\r\n";
	exit;
}

sub sigexit_smtpdummy_forceclosed {
	write_log LOG_LV_INFO,LOG_NG,"SMTP","connection closed (force)";
	
	# add ver1.2
	if ( ${Mode_SmtpTimeout_Deny} == 1 ) {
		if ( get_hostinfo ${Path_Greylist},${Objectipaddr} ){
			delete_hostinfo_greylist ${Path_Greylist},${Objectipaddr};
		}
	}
	
	print "421 ${Domain} timeout\r\n";
	exit;
}

sub smtp_deny_fail {
	print "450 Because your smtp host was registered with my blacklist, we denied connection from your host.\r\n";
	write_log LOG_LV_DEBUG,LOG_INFO,"SMTP","connection closed (deny)";
	exit;
}

sub get_reverse_hostname {
	my ${hostname}="";
	my ${hostname_rev}="";
	
	( ${hostname} ) = @_;
	
	${hostname_rev} = reverse ${hostname};
	
	if ( ${hostname_rev} =~ /^([^\.]+)\.([^\.]*)\.([^\.]*)\.([^\.]*)/i ){
		return ($1,$2,$3,$4);
	}elsif ( ${hostname_rev} =~ /^([^\.]+)\.([^\.]*)\.([^\.]*)/i ) {
		return ($1,$2,$3,"-");
	}elsif ( ${hostname_rev} =~ /^([^\.]+)\.([^\.]*)/i ) {
		return ($1,$2,"-","-");
	}elsif ( ${hostname_rev} =~ /^([^\.]+)/i ) {
		return ($1,"-","-","-");
	}	
	return undef;
}

sub check_helohost_connecthost_sub {
	my ${helohostname}="";
	my ${connecthostname}="";
	
	my @{helohostname_rev};
	my @{connecthostname_rev};
	
	(  ${helohostname},${connecthostname} ) = @_;
	
	@{helohostname_rev} = get_reverse_hostname ${helohostname};
	@{connecthostname_rev} = get_reverse_hostname ${connecthostname};
	
	if ( ${Mode_Compare_Hostname_Octetlevel} >= 1 ) {
		if ( ${helohostname_rev}[0] ne ${connecthostname_rev}[0] ) {
			return "NG";
		}
	}
	if ( ${Mode_Compare_Hostname_Octetlevel} >= 2 ) {
		if ( ${helohostname_rev}[1] ne ${connecthostname_rev}[1] ) {
			return "NG";
		}
	}
	
	# Probably there is not a meaning to carry out
	if ( ${Mode_Compare_Hostname_Octetlevel} >= 3 ) {
		if ( ${helohostname_rev}[2] ne ${connecthostname_rev}[2] ) {
			return "NG";
		}
	}
	
	# modify ver1.2.4
	# return ;
	return undef;
}

sub check_helohost_connecthost {
	my ${helohostname}="";
	my ${connecthostname}="";
	
	(  ${helohostname},${connecthostname} ) = @_;
	
	# ver1.1 add
	if ( ${Mode_Check_RemotoIp_HeloIp} == 1 ) {
		if ( ${helohostname} =~ /^(\[*\d+\.\d+\.\d+\.\d+\]*|[0-9A-Fa-f]{1,4}:)/ ) {
			if ( ${helohostname} ne ${Remoteip} ) {
				# It is deleted the information of this IP address by greylist
				write_log LOG_LV_INFO,LOG_NG,"helohost_connecthost_check","different from HELO host in connection host - HELO:${helohostname}";
				delete_hostinfo_greylist ${Path_Greylist},${Objectipaddr};
				smtp_deny_fail;
			}
		}
	}
	
	# ver1.1 add
	if ( ( ${Mode_Compare_Hostname_Octetlevel} > 0 ) and
	     ( ${Remoteip} ne ${Remotehostip} ) ) {
	 
		if ( check_helohost_connecthost_sub ${helohostname},${connecthostname} ) {
			# It is deleted the information of this IP address by greylist
			write_log LOG_LV_INFO,LOG_NG,"helohost_connecthost_check","different from HELO host in connection host - HELO:${helohostname}";
			delete_hostinfo_greylist ${Path_Greylist},${Objectipaddr};
			smtp_deny_fail;
		}
		
	}
}

sub smtp_dummy {
	my ${failmsg}="";
	my ${inputline}="";
	my ${smtpcmd}="";
	my ${smtpvalue}="";
	my ${addr}="";
	my ${replymsg}="";
	my ${count_smtpcmd}=0;
	my ${count_smtpcmd_unknown}=0; # Add ver1.2

	( ${failmsg} ) = @_;

	print "220 ${Msg_Connection}\r\n";

	# Setting of the trap handling of in time-out
	$SIG{'ALRM'} = "sigexit_smtpdummy_timeout";
	$SIG{'INT'} = $SIG{'TERM'} = $SIG{'HUP'} = $SIG{'PIPE'} = $SIG{'QUIT'} = "sigexit_smtpdummy_forceclosed";

	${count_smtpcmd}=0;
	${count_smtpcmd_unknown}= 0; # add ver1.2
	# Time limit setting（It is an anti-silence measure from the beginning）
	alarm ${Sec_Smtptimeout};
	while ( ${inputline} = <STDIN> ) {
		alarm 0;
		# reply delay
		sleep ${Sec_Smtpcmdwait};
		# Time limit setting
		alarm ${Sec_Smtptimeout};

		# cut off a new line
		${inputline} =~ s/[\r\n]//g;
		${inputline} =~ s/^\s+//g;

		if ( ${inputline} eq "" ) {
			print "500 not known\r\n";
			next;
		}

		# disintegrate in a command and a value
		${inputline} =~ /^(\S+)\s*(.*)$/;
		${smtpcmd}=$1;
		${smtpvalue}=$2;


		${replymsg}="";
		if ( ${smtpcmd} =~ /(HELO|EHLO)/i ) {
			if ( defined ${smtpvalue} ) {
				if ( ${Mode_Badhelohost} == 1 ) {
					# ver1.1 change
					#if ( check_hostname ${Path_Blacklist_Helohost},${smtpvalue} ) {
					if ( check_helohostname ${Path_Blacklist_Helohost},${smtpvalue},${Remotehostip} ) {
						write_log LOG_LV_INFO,LOG_NG,"helohostcheck","bad remote helo : ${Remotehostip} -> ${smtpvalue}";
						delete_hostinfo_greylist ${Path_Greylist},${Objectipaddr};
						smtp_deny_fail;
					}else{
						write_log LOG_LV_DEBUG,LOG_INFO,"helohostcheck","remote helo : ${Remotehostip} -> ${smtpvalue}";
					}
				}else{
					write_log LOG_LV_DEBUG,LOG_INFO,"remote helo","${Remotehostip} -> ${smtpvalue}";
				}
				
				# ver1.1 change
				#if ( ( ${Mode_Compare_Hostname_Octetlevel} > 0 ) and
				#     ( ${Remoteip} ne ${Remotehostip} ) ) {
				#	check_helohost_connecthost ${smtpvalue},${Remotehostip};
				#}
				check_helohost_connecthost ${smtpvalue},${Remotehostip};
			}else{
				write_log LOG_LV_DEBUG,LOG_INFO,"remote helo","${Remotehostip} -> --";
			}

			# Response message
			if ( ( ${Mode_Escape_SmtpAuth} == 1) and
			     ( ${smtpcmd} =~ /EHLO/i ))
			{
				${replymsg}="250-${Domain}\r\n250 AUTH PLAIN LOGIN";
			}else{
				${replymsg}="250 hello";			
			}

		}elsif (${smtpcmd} =~ /MAIL/i){
			if ( ${smtpvalue} eq "" ) {
				${addr}="?";
			}else{
				${addr}=${smtpvalue};
			}
			${addr} =~ s/^from:? *//i;

			# Response message
			${replymsg}="250 Mail from ${addr}";
		}elsif (${smtpcmd} =~ /RCPT/i){
			if ( ${smtpvalue} eq "" ) {
				${addr}="?";
			}else{
				${addr}=${smtpvalue};
			}
			${addr} =~ s/^to:? *//i;

			# Response message
			${replymsg}="450 Rcpt to ${addr} - ${failmsg}";
		}elsif (${smtpcmd} =~ /DATA/i){
			# Response message
			${replymsg}="451 ${failmsg}";
		}elsif (${smtpcmd} =~ /RSET/i){
			# Response message
			${replymsg}="250 ready";
		}elsif (${smtpcmd} =~ /VRFY/i){
			# Response message
			${replymsg}="502 not implemented";
		}elsif (${smtpcmd} =~ /EXPN/i){
			# Response message
			${replymsg}="502 not implemented";
		}elsif (${smtpcmd} =~ /HELP/i){
			# Response message
			${replymsg}="502 not implemented";
		}elsif (${smtpcmd} =~ /NOOP/i){
			# Response message
			${replymsg}="502 not implemented";
		}elsif (${smtpcmd} =~ /QUIT/i){
			# Response message
			${replymsg}="221 ${Domain} Bye";
		}elsif (${smtpcmd} =~ /AUTH/i){
			write_log LOG_LV_DEBUG,LOG_INFO,"smtp_auth","smtp_auth request";
			if ( ${Mode_Escape_SmtpAuth} == 1 ) {
				write_log LOG_LV_INFO,LOG_OK,"smtp_auth","request";
				
				set_create_access_time_soonaccept_greylist ${Path_Greylist},${Objectipaddr};
				# Response message
				${replymsg}="454 smtp-auth temporary error. Please try again.";

			}else{
				# Response message
				${replymsg}="502 not implemented";
			}
		# add ver1.2
		}elsif (${smtpcmd} =~ /ATRN/i){
			# Response message
			${replymsg}="502 not implemented";
		}elsif (${smtpcmd} =~ /BDAT/i){
			# Response message
			${replymsg}="502 not implemented";
		}elsif (${smtpcmd} =~ /ETRN/i){
			# Response message
			${replymsg}="502 not implemented";
		}elsif (${smtpcmd} =~ /SAML/i){
			# Response message
			${replymsg}="502 not implemented";
		}elsif (${smtpcmd} =~ /SEND/i){
			# Response message
			${replymsg}="502 not implemented";
		}elsif (${smtpcmd} =~ /SOML/i){
			# Response message
			${replymsg}="502 not implemented";
		}elsif (${smtpcmd} =~ /STARTTLS/i){
			# Response message
			${replymsg}="502 not implemented";
		}elsif (${smtpcmd} =~ /TURN/i){
			# Response message
			${replymsg}="502 not implemented";
		}else{
			# Response message
			${replymsg}="500 ${smtpcmd} not known";
			${count_smtpcmd_unknown}++;
		}

		print "${replymsg}\r\n";
		write_log LOG_LV_DEBUG,LOG_INFO,"SMTP","${inputline} => ${replymsg}";

		${count_smtpcmd}++;

		if (${smtpcmd} =~ /QUIT/i){
			last ;
		}

	}

	if ( ${count_smtpcmd} == 0 ) {
		write_log LOG_LV_INFO,LOG_NG,"tarpit","no smtp command - tarpit timeout ";
		delete_hostinfo_greylist ${Path_Greylist},${Objectipaddr};
		smtp_deny_fail;
	}
	
	# add ver1.2
	if ( ${count_smtpcmd} <= ${Threshold_Count_SmtpCommand} ) {
		write_log LOG_LV_INFO,LOG_NG,"SMTP","too little number of the SMTP command";
		delete_hostinfo_greylist ${Path_Greylist},${Objectipaddr};
		smtp_deny_fail;	
	}
	
	# add ver1.2
	if ( ${Threshold_Count_UnknownSmtpCmd} != 0 ) {
		if ( ${count_smtpcmd_unknown} >= ${Threshold_Count_UnknownSmtpCmd} ) {
			write_log LOG_LV_INFO,LOG_NG,"SMTP","it is much number of unknown SMTP command";
			delete_hostinfo_greylist ${Path_Greylist},${Objectipaddr};
			smtp_deny_fail;
		}
	}

	write_log LOG_LV_DEBUG,LOG_INFO,"SMTP","connection closed (quit)";
	
	alarm 0;
	$SIG{'INT'} = $SIG{'TERM'} = $SIG{'HUP'} = $SIG{'ALRM'} = $SIG{'PIPE'} = $SIG{'QUIT'} ="";
	exit;
}

sub check_greylist_ng{
	my ${sec_interval_lasttime};
	my ${count_comesoon} = 0;
	# It deletes expired host information from greylist.
	delete_expirationhostinfo_greylist_main;


	if ( ! get_hostinfo ${Path_Greylist},${Objectipaddr} ) {
		# When there is not the information of this IP address in greylist
		write_log LOG_LV_INFO,LOG_INFO,"greylisting","noregistered host";
		
		# add -- tarpit
		if ( ${Mode_Tarpit} == 1 ) {
			tarpit ${Remoteip};
		}

		# set greylist
		set_access_time_now_greylist ${Path_Greylist},${Objectipaddr};

		# modify ver1.3.0
		#smtp_dummy "Temporary local problem. Please try later.";
		if ( ${Mode_PrintMsg_AccurateReason} == 1 ) {
			smtp_dummy "your host registered greylist. Please try later.";
		}else {
			smtp_dummy "Temporary local problem. Please try later.";
		}

	}else{
		# When there is the information of this IP address in greylist
		
		( ${Timesirial_Create} , ${Timesirial_Access}, ${count_comesoon} ) = get_create_access_time_greylist ${Path_Greylist},${Objectipaddr};
		
		# add ver1.2.1 if only
		if ( defined ${Timesirial_Create} and ${Timesirial_Create} ne '' ) {
			${sec_interval_lasttime} = time - ${Timesirial_Create};
			
			if ( ( ${Threshold_Count_Comesoon} > 0 ) and
			     ( ${Threshold_Count_Comesoon} <= ${count_comesoon} ) ){
			 	# When, more than the number of times that he appointed, 
			 	# this host tried connection before suppression time
				write_log LOG_LV_INFO,LOG_NG,"greylisting","The host who came more than ${Threshold_Count_Comesoon} times soon. ${count_comesoon} times";
				# count up comesoon count
				countup_comesooncount_greylist ${Path_Greylist},${Objectipaddr};
				
				# modify ver1.3.0
				#smtp_dummy "Temporary local problem. Please try later.";
				if ( ${Mode_PrintMsg_AccurateReason} == 1 ) {
					smtp_dummy "your host registered greylist. Please try later.";
				}else{
					smtp_dummy "Temporary local problem. Please try later.";
				}
			
			}
			
			if ( ${sec_interval_lasttime} < ${Sec_Suppression_Connect} ) {
				
				# modify ver 1.2.4 LOG_NG
				write_log LOG_LV_INFO,LOG_NG,"greylisting","back too soon(${sec_interval_lasttime} sec from last time). temp error again";
				# count up comesoon count
				countup_comesooncount_greylist ${Path_Greylist},${Objectipaddr};
				
				# modify ver1.3.0
				#smtp_dummy "Temporary local problem. Please try later.";
				if ( ${Mode_PrintMsg_AccurateReason} == 1 ) {
					smtp_dummy "your host registered greylist. Please try later.";
				}else{
					smtp_dummy "Temporary local problem. Please try later.";
				}
			}
			# access the check file as it's a good IP
			set_access_time_now_greylist ${Path_Greylist},${Objectipaddr};
			
			write_log LOG_LV_DEBUG,LOG_INFO,"greylisting","registered host ok (${sec_interval_lasttime} sec from last time)";
		}else{
			# modify ver1.2.5 
			# write_log LOG_LV_DEBUG,LOG_WARN,"greylisting","invalid format";
			write_log LOG_LV_DEBUG,LOG_NG,"greylisting","It may have been deleted from greylist";
			# add ver1.2.5 
			smtp_dummy "Temporary local problem. Please try later.";
		}
	}

}

sub check_nohit_s25rlist_ok {
	# add -- Continue - s25r list
	if ( defined $ENV{TCPREMOTEHOST} and $ENV{TCPREMOTEHOST} ne '' ) {
		if ( ( ! check_hostname ${Path_S25rlist_Hostname},$ENV{TCPREMOTEHOST} )
		     and (-f ${Path_S25rlist_Hostname})                                ) {
			write_log LOG_LV_INFO,LOG_OK,"S25R","nomatch s25rlist $ENV{TCPREMOTEHOST}";
			exec_next;
		}
	}
}

sub check_hit_blacklist_ng {
	# add -- NG - blacklisted IPs
	if ( get_hostinfo ${Path_Blacklist},${Objectipaddr} ) {
		write_log LOG_LV_INFO,LOG_NG,"blacklist","blacklisted ip";
		smtp_deny_fail;
	}
	# add -- NG - blacklisted hostnaem
	if ( check_hostname ${Path_Blacklist_Hostname},${Remotehostip} ) {
		write_log LOG_LV_INFO,LOG_NG,"blacklist","blacklisted hostname";
		smtp_deny_fail;
	}
}

sub check_hit_whitelist_ok {
	# ok - whitelisted IPs
	if ( get_hostinfo ${Path_Whitelist},${Objectipaddr} ) {
		write_log LOG_LV_INFO,LOG_OK,"whitelist","whitelisted ip";
		exec_next;
	}

	# add -- OK - whitelisted hostname
	if ( check_hostname ${Path_Whitelist_Hostname},${Remotehostip} ) {
		write_log LOG_LV_INFO,LOG_OK,"whitelist","whitelisted hostname";
		exec_next;
	}
}


sub check_remoteip_main {
	if ( ! defined $ENV{TCPREMOTEIP} )
	{
		write_log LOG_LV_FORCE,LOG_ERR,"-","ENV{TCPREMOEIP} not set!";
		smtp_dummy "Cannot find remote IP";
	}

	if ( ! $ENV{TCPREMOTEIP} =~ /^(\d+\.\d+\.\d+\.\d+|[0-9A-Fa-f]{1,4}:)/ )
	{
		write_log LOG_LV_FORCE,LOG_ERR,"-","bad format : ENV{TCPREMOEIP} = $ENV{TCPREMOTEIP}";
		smtp_dummy "Bad format for remote IP";
	}
}

#---------------------------------------#
# main proc     (do not edit)           #
#---------------------------------------#

# inspection of the IP address
check_remoteip_main;

${Remoteip} = $ENV{TCPREMOTEIP};

# set hostname or ipaddr
if (  defined $ENV{TCPREMOTEHOST} and $ENV{TCPREMOTEHOST} ne '' ) {
	if ( $ENV{TCPREMOTEHOST} eq "unknown" ) {
		${Remotehostip}=$ENV{TCPREMOTEIP};
	}else{
		${Remotehostip}=$ENV{TCPREMOTEHOST};
	}
}else {
	${Remotehostip}=$ENV{TCPREMOTEIP};
}

write_log LOG_LV_DEBUG,LOG_INFO,"-","start";

# get  an object IP address to use in greylist
${Objectipaddr} = get_objectipaddr  ${Remoteip};

# do not carry out a check to a relay client
if ( defined $ENV{RELAYCLIENT} )
{
	write_log LOG_LV_INFO,LOG_OK,"-","Relay client accepted";
	exec_next;
}

# add ver1.3.0
# escape SMTP-AUTH connection. （only qmail-vida)
if ( ( ${Mode_Escape_SmtpAuth_ForVida} == 1) and 
     ( defined $ENV{USER} ) ) {
	write_log LOG_LV_INFO,LOG_OK,"-","SMTP AUTHed client accepted";
	exec_next;
}

# whitelist
check_hit_whitelist_ok;

if ( ${Mode_Blacklist} == 2 ) {
	# blacklist
	check_hit_blacklist_ng;
}

if ( ${Mode_S25rlist} == 1 ) {
	# s25rlist
	check_nohit_s25rlist_ok;
}

if ( ${Mode_Blacklist} == 1 ) {
	# blacklist
	check_hit_blacklist_ng;
}

if ( ${Mode_Tarpit} == 2 ) {
	# tarpit (force all)
	tarpit ${Remoteip};
}

if ( ${Mode_Greylist} == 1 ) {
	# greylist
	check_greylist_ng;
}

# Because it passed all the inspections, it permits. 
write_log LOG_LV_INFO,LOG_OK,"-","accepting";

# cooperates qmail-smtpd
exec_next;

